/* ©  Copyright Steve Simon, 2013 */
#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <auth.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include <json.h>
#include "jira.h"


static void
links(Node *n, Jvalue *v)
{
	int i;
	char *p;
	Node *c;

	for(i = 0; (p = Jstrval(v, "fields.links.value[%d].issueKey", i)) != nil; i++){
		c = subdir(&n->child, nil, nil, p);
		if((p = Jstrval(v, "fields.tasks.value[%d].type.name", i)) != nil)
			subtext(&c->child, "name", p);
		if((p = Jstrval(v, "fields.tasks.value[%d].type.direction", i)) != nil)
			subtext(&c->child, "direction", p);
		if((p = Jstrval(v, "fields.tasks.value[%d].type.description", i)) != nil)
			subtext(&c->child, "description", p);
	}
}

static void
subtasks(Node *n, Jvalue *v)
{
	int i;
	char *p;
	Node *c;

	for(i = 0; (p = Jstrval(v, "fields.sub-tasks.value[%d].key", i)) != nil; i++){
		c = subdir(&n->child, nil, nil, p);
		if((p = Jstrval(v, "fields.tasks.value[%d].type.name", i)) != nil)
			subtext(&c->child, "name", p);
		if((p = Jstrval(v, "fields.tasks.value[%d].type.direction", i)) != nil)
			subtext(&c->child, "direction", p);
	}
}


static void
attachments(Node *n, Jvalue *v, Sess *s)
{
	int i;
	long created;
	vlong length;
	char *p, *author, *file, *content;

	for(i = 0; (file = Jstrval(v, "fields.attachment.value[%d].filename", i)) != nil; i++){

		if((author = Jstrval(v, "fields.attachment.value[%d].author.name", i)) != nil)
			adduser(s, author);
		else
			author = "unknown";

		if((p = Jstrval(v, "fields.attachment.value[%d].created", i)) != nil)
			created = iso2sec(p);
		else
			created = time(nil);

		if((p = Jstrval(v, "fields.attachment.value[%d].size", i)) != nil)
			length = strtoull(p, nil, 10);
		else
			length = 0;

		if((content = Jstrval(v, "fields.attachment.value[%d].content", i)) != nil)
			fetchtext(&n->child, file, author, nil, author, created, length, content);
	}
}

static void
worklog(Node *n, Jvalue *v, Sess *s)
{
	int i;
	Node *c;
	char *p, dir[16];

	for(i = 0; (p = Jstrval(v, "fields.worklog.value[%d].author.name", i)) != nil; i++){
		snprint(dir, sizeof(dir), "%d", i);
		c = subdir(&n->child, dir, nil, p);

		subtext(&c->child, "author", p);
		adduser(s, p);

		if((p = Jstrval(v, "fields.worklog.value[%d].updateAuthor.name", i)) != nil)
			subtext(&c->child, "updater", p);
		adduser(s, p);

		if((p = Jstrval(v, "fields.worklog.value[%d].comment", i)) != nil)
			subtext(&c->child, "comment", p);
		if((p = Jstrval(v, "fields.worklog.value[%d].started", i)) != nil)
			subtext(&c->child, "started", ctime(iso2sec(p)));
		if((p = Jstrval(v, "fields.worklog.value[%d].minutesSpent", i)) != nil)
			subtext(&c->child, "minuites", p);
	}
}

static void
timetracking(Node **np, Jvalue *v)
{
	Fmt f;
	char *p, *buf;

	fmtstrinit(&f);
	if((p = Jstrval(v, "fields.timetracking.value.timeoriginalestimate")) != nil)
		fmtprint(&f, "%-16q %q\n", "originalestimate", p);
	if((p = Jstrval(v, "fields.timetracking.value.timeestimate")) != nil)
		fmtprint(&f, "%-16q %q\n", "estimate", p);
	if((p = Jstrval(v, "fields.timetracking.value.timespent")) != nil)
		fmtprint(&f, "%-16q %q\n", "spent", p);
	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	if(strlen(buf) > 0)
		subtext(np, "timetracking", buf);
	free(buf);
}

static void
transitions(Node **n, Sess *s, char *issue, Jvalue *projvals)
{
	Fmt f;
	int i;
	Jmap *m;
	Jvalue *v;
	char *p, *buf, *req, *name, *field, *type, *sep;

	if((v = get(s, "rest/api/%s/issue/%U/transitions", s->api, issue)) == nil){
		warn("%s cannot get valid transition list\n", issue);
		return;
	}

	fmtstrinit(&f);
	fmtprint(&f, "%-4q %-24q %-24s %-16s %s\n", "#Id", "Transition", "Field", "Required", "Type");

	for(m = Jmapfirst(v, ""); m; m = m->next){
		sep = " ";

		if((name = Jstrval(v, "%s.name", m->name)) == nil)
			name = "";
		fmtprint(&f, "%-4q %-24q", m->name, name);

		for(i = 0; (field = Jstrval(v, "%s.fields[%d].id", m->name, i)); i++){
			if((req = Jstrval(v, "%s.fields[%d].required", m->name, i)) == nil)
				continue;

			if(strncmp(field, "customfield_", 12) == 0)				/* map custom field names to human */
				if((p = Jstrval(projvals, "fields.%s.name", field)) != nil)
					field = p;

			if((type = Jstrval(v, "%s.fields[%d].type", m->name, i)) == nil)
				type = "";
			if((p = strrchr(type, ':')) != nil)			/* custom fields */
				type = p+1;
			if((p = strrchr(type, '.')) != nil)			/* builtin fields */
				type = p+1;

			if(strcmp(req, "true") == 0)
				req = "mandatory";
			else
				req = "optional";

			fmtprint(&f, "%s%-24q %-16s %s\n", sep, field, req, type);
			sep = "                              ";
		}
		fmtprint(&f, "\n");

	}

	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	writehook(subtext(n, "transition", buf), add_transition, issue);
	free(buf);

	Jfree(v);
}



static void
comments(Node *n, Jvalue *v, Sess *s)
{
	int i;
	long created, updated;
	char *p, *author, *body, file[16];

	for(i = 0; (body = Jstrval(v, "fields.comment.value[%d].body", i)) != nil; i++){

		if((author = Jstrval(v, "fields.comment.value[%d].author.name", i)) == nil)
			author = "unknown";
		else
			adduser(s, author);

		if((p = Jstrval(v, "fields.comment.value[%d].created", i)) != nil)
			created = iso2sec(p);
		else
			created = time(nil);

		if((p = Jstrval(v, "fields.comment.value[%d].updated", i)) != nil)
			updated = iso2sec(p);
		else
			updated = time(nil);

		snprint(file, sizeof(file), "%d", i);
		fulltext(&n->child, file, author, nil, author, created, updated, body);
	}
}

Node *
issueinfo_v4(Node *n, Sess *s, char *issue)
{
	Jvalue *v;
	char *p, *uid;

	if((v = get(s, "rest/api/%s/issue/%U", s->api, issue)) == nil){
		warn("%s cannot get issue detail\n", issue);
		return n;
	}

	if((uid = Jstrval(v, "fields.reporter.value.name")) != nil){
		adduser(s, uid);
		fulltext(&n, "reporter", uid, nil, uid, 0, 0, uid);
	}
	if((p = Jstrval(v, "fields.status.value.name")) != nil)
		fulltext(&n, "status", uid, nil, uid, 0, 0, p);
	if((p = Jstrval(v, "fields.resolution.value.name")) != nil)
		fulltext(&n, "resolution", uid, nil, uid, 0, 0, p);
	if((p = Jstrval(v, "fields.created")) != nil)
		fulltext(&n, "created", uid, nil, uid, 0, 0, ctime(iso2sec(p)));
	if((p = Jstrval(v, "fields.updated")) != nil)
		fulltext(&n, "updated", uid, nil, uid, 0, 0, ctime(iso2sec(p)));
	if((p = Jstrval(v, "fields.resolutiondate.value")) != nil)
		fulltext(&n, "resolved", uid, nil, uid, 0, 0, ctime(iso2sec(p)));
	if((p = Jstrval(v, "fields.environment.value")) != nil)
		fulltext(&n, "environment", uid, nil, uid, 0, 0, p);
	if((p = Jstrval(v, "FixVersions.value.name")) != nil)
		fulltext(&n, "fix-version", uid, nil, uid, 0, 0, p);

	if((p = Jstrval(v, "fields.assignee.value.name")) == nil)
		p = "";
	else
		adduser(s, p);
	writehook(fulltext(&n, "assignee", uid, nil, uid, 0, 0, p), upd_assignee, issue);

	if((p = Jstrval(v, "fields.issuetype.value.name")) == nil)
		p = "";
	writehook(fulltext(&n, "issuetype", uid, nil, uid, 0, 0, p), upd_issuetype, issue);

	if((p = Jstrval(v, "fields.summary.value")) == nil)
		p = "";
	writehook(fulltext(&n, "summary", uid, nil, uid, 0, 0, p), upd_summary, issue);

	if((p = Jstrval(v, "fields.description.value")) == nil)
		p = "";
	writehook(fulltext(&n, "description", uid, nil, uid, 0, 0, p), upd_description, issue);

	timetracking(&n, v);
	transitions(&n, s, issue, v);
	
	worklog(subdir(&n, nil, nil, "worklog"), v, s);
	comments(subdir(&n, nil, nil, "comments"), v, s);
	attachments(subdir(&n, nil, nil, "attachments"), v, s);
	subtasks(subdir(&n, nil, nil, "subtasks"), v);
	links(subdir(&n, nil, nil, "links"), v);

	Jfree(v);
	return n;
}

