/* ©  Copyright Steve Simon, 2013 */
#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <auth.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include <json.h>
#include "jira.h"



/***************************************************************/

static void
links(Node *n, Jvalue *v)	// FIXME: not right at all...
{
	int i;
	char *p;

	for(i = 0; (p = Jstrval(v, "fields.issuelinks[%d].type.name", i)) != nil; i++){
			subtext(&n->child, "name", p);
		if((p = Jstrval(v, "fields.issuelinks[%d].outward", i)) != nil)
			subtext(&n->child, "outward", p);
		if((p = Jstrval(v, "fields.issuelinks[%d].type.direction", i)) != nil)
			subtext(&n->child, "direction", p);
		if((p = Jstrval(v, "fields.issuelinks[%d].type.description", i)) != nil)
			subtext(&n->child, "description", p);
	}
}

static void
subtasks(Node *n, Jvalue *v)
{
	int i;
	char *p;
	Node *c;

	for(i = 0; (p = Jstrval(v, "fields.sub-tasks[%d].key", i)) != nil; i++){
		c = subdir(&n->child, nil, nil, p);
		if((p = Jstrval(v, "fields.tasks[%d].type.name", i)) != nil)
			subtext(&c->child, "name", p);
		if((p = Jstrval(v, "fields.tasks[%d].type.direction", i)) != nil)
			subtext(&c->child, "direction", p);
	}
}


static void
attachments(Node *n, Jvalue *v, Sess *s)
{
	int i;
	vlong length;
	long created;
	char *p, *author, *file, *content;

	for(i = 0; (file = Jstrval(v, "fields.attachment[%d].filename", i)) != nil; i++){

		if((author = Jstrval(v, "fields.attachment[%d].author.name", i)) != nil)
			adduser(s, author);
		else
			author = "unknown";

		if((p = Jstrval(v, "fields.attachment[%d].created", i)) != nil)
			created = iso2sec(p);
		else
			created = time(nil);

		if((p = Jstrval(v, "fields.attachment[%d].size", i)) != nil)
			length = strtoull(p, nil, 10);
		else
			length = 0;

		if((content = Jstrval(v, "fields.attachment[%d].content", i)) != nil)
			fetchtext(&n->child, file, author, nil, author, created, length, content);
	}
}

static void
watchers(Node **np, Jvalue *v)
{
	int i;
	Fmt f;
	char *p, *buf;

	fmtstrinit(&f);
	for(i = 0; (p = Jstrval(v, "watcher.watchers[%d].name", i)) != nil; i++)
		fmtprint(&f, "%s\n", p);
	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	if(strlen(buf) > 0)
		subtext(np, "watchers", buf);
	free(buf);
}

static void
versions(Node **np, Jvalue *v, char *file, char *field)
{
	int i;
	Fmt f;
	char *p, *buf;

	fmtstrinit(&f);
	fmtprint(&f, "%-16s %-10s %-10s %-12s %s\n", "#Name", "Archived", "Released", "Rel-date", "Description");
	for(i = 0; (p = Jstrval(v, "fields.%s[%d].name", field, i)) != nil; i++){
		fmtprint(&f, "%-16q ", p);

		if((p = Jstrval(v, "fields.%s[%d].archived", field, i)) != nil)
			fmtprint(&f, "%-10q ", p);
		else
			fmtprint(&f, "%-10s ", "-");
		if((p = Jstrval(v, "fields.%s[%d].released", field, i)) != nil)
			fmtprint(&f, "%-10q ", p);
		else
			fmtprint(&f, "%-10s ", "-");
		if((p = Jstrval(v, "fields.%s[%d].releaseDate", field, i)) != nil)
			fmtprint(&f, "%-12q ", p);
		else
			fmtprint(&f, "%-12s ", "-");
		if((p = Jstrval(v, "fields.%s[%d].description", field, i)) != nil)
			fmtprint(&f, "%s\n", p);
		else
			fmtprint(&f, "%s\n", "-");
	}
	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	if(strlen(buf) > 0)
		subtext(np, file, buf);
	free(buf);
}

static void
components(Node **np, Jvalue *v)
{
	int i;
	Fmt f;
	char *p, *buf;

	fmtstrinit(&f);
	for(i = 0; (p = Jstrval(v, "fields.components[%d].name", i)) != nil; i++)
		fmtprint(&f, "%s\n", p);
	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	if(strlen(buf) > 0)
		subtext(np, "components", buf);
	free(buf);
}

static void
progress(Node **np, Jvalue *v, char *path, char *file)
{
	Fmt f;
	char *p, *buf;

	fmtstrinit(&f);
	if((p = Jstrval(v, "%s.progress", path)) != nil)
		fmtprint(&f, "progress	%s\n", p);
	if((p = Jstrval(v, "%s.total", path)) != nil)
		fmtprint(&f, "total	%s\n", p);
	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	if(strlen(buf) > 0)
		subtext(np, file, buf);
	free(buf);
}

static void
worklog(Node *n, Jvalue *v, Sess *s)
{
	int i;
	Node *c;
	char *p, *q, dir[16], buf[128];

	for(i = 0; (p = Jstrval(v, "fields.worklog[%d].author.name", i)) != nil; i++){
		snprint(dir, sizeof(dir), "%d", i);
		c = subdir(&n->child, dir, nil, p);

		subtext(&c->child, "author", p);
		adduser(s, p);

		if((p = Jstrval(v, "fields.worklog[%d].updateAuthor.name", i)) != nil)
			subtext(&c->child, "updater", p);
		adduser(s, p);

		if((p = Jstrval(v, "fields.worklog[%d].comment", i)) != nil)
			subtext(&c->child, "comment", p);
		if((p = Jstrval(v, "fields.worklog[%d].started", i)) != nil)
			subtext(&c->child, "started", ctime(iso2sec(p)));
		if((p = Jstrval(v, "fields.worklog[%d].timeSpent", i)) != nil){
			if((q = Jstrval(v, "fields.worklog[%d].timeSpentSeconds", i)) != nil)
				snprint(buf, sizeof(buf), "%s %s", p, q);
			else
				snprint(buf, sizeof(buf), "%s", p);
			subtext(&c->child, "minuites", p);
		}
	}
}

static void
timetracking(Node **np, Jvalue *v)
{
	Fmt f;
	char *p, *q, *buf;

	fmtstrinit(&f);
	if((p = Jstrval(v, "fields.timetracking.timeoriginalestimate")) != nil)
		if((q = Jstrval(v, "fields.timetracking.originalEstimateSeconds")) != nil)
			fmtprint(&f, "%-16q %q %q\n", "originalEstimate", p, q);
		else
			fmtprint(&f, "%-16q %q\n", "originalEstimate", p);

	if((p = Jstrval(v, "fields.timetracking.remainingEstimate")) != nil)
		if((q = Jstrval(v, "fields.timetracking.remainingEstimateSeconds")) != nil)
			fmtprint(&f, "%-16q %q %q\n", "remainingEstimate", p, q);
		else
			fmtprint(&f, "%-16q %q\n", "remainingEstimate", p);

	if((p = Jstrval(v, "fields.timetracking.timespent")) != nil)
		if((q = Jstrval(v, "fields.timetracking.timespentSeconds")) != nil)
			fmtprint(&f, "%-16q %q %q\n", "timespent", p, q);
		else
			fmtprint(&f, "%-16q %q\n", "timespent", p);

	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	if(strlen(buf) > 0)
		subtext(np, "timetracking", buf);
	free(buf);
}

static void
transitions(Node **np, Sess *s, char *issue)
{
	Fmt f;
	int i, j;
	Jmap *m;
	Jvalue *v;
	char *p, *buf, *sep, *name, *id, *type, *field, *req, *value;

	if((v = get(s, "rest/api/%s/issue/%U/transitions?expand=transitions.fields", s->api, issue)) == nil){
		warn("%s cannot get valid transition list - %r\n", issue);
		return;
	}

	fmtstrinit(&f);
	fmtprint(&f, "%-4q %-24q %-24s %-16s %s\n", "#Id", "Transition", "Field", "Required", "Type");
	for(i = 0; (name = Jstrval(v, "transitions[%d].name", i)) != nil; i++){
		sep = " ";
		if((id = Jstrval(v, "transitions[%d].id", i)) == nil)
			continue;
		fmtprint(&f, "%-4q %-24q", id, name);

		for(m = Jmapfirst(v, "transitions[%d].fields", i); m; m = m->next){
			field = m->name;
			if((req = Jstrval(v, "transitions[%d].fields.%s.required", i, m->name)) == nil)
				req = "";

			if(strcmp(req, "true") != 0)	/* only show mandatory fields */
				continue;

			/* custom fields */
			if((p = Jstrval(v, "transitions[%d].fields.%s.schema.custom", i, m->name)) != nil){
				type = p;
				if((p = strrchr(type, ':')) != nil)
					type = p+1;

				if((p = Jstrval(v, "transitions[%d].fields.%s.name", i, m->name)) != nil)
					field = p;
			}
			else
			/* builtin fields */
			if((p = Jstrval(v, "transitions[%d].fields.%s.schema.type", i, m->name)) != nil)
				type = p;
			else
				type = "";

			if(strcmp(req, "true") == 0)
				req = "mandatory";
			else
				req = "optional";

			fmtprint(&f, "%s%-24q %-16s %-16s ", sep, field, req, type);

			/* custom fields */
			for(j = 0; (value = Jstrval(v, "transitions[%d].fields.%s.allowedValues[%d].value", i, m->name, j)) != nil; j++){
				if(j)
					fmtprint(&f, "|%s", value);
				else
					fmtprint(&f, "%s", value);
			}

			/* builtin fields */
			for(j = 0; (value = Jstrval(v, "transitions[%d].fields.%s.allowedValues[%d].name", i, m->name, j)) != nil; j++){
				if(j)
					fmtprint(&f, "|%s", value);
				else
					fmtprint(&f, "%s", value);
			}

			fmtprint(&f, "\n");
			sep = "                              ";
		}
		fmtprint(&f, "\n");
	}

	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	writehook(subtext(np, "transitions", buf), add_transition, issue);

	free(buf);
	Jfree(v);
}

static void
comments(Node *n, Jvalue *v, Sess *s)
{
	int i;
	long created, updated;
	char *p, *uid, *muid, *body, file[16];

	for(i = 0; (body = Jstrval(v, "fields.comment.comments[%d].body", i)) != nil; i++){

		if((uid = Jstrval(v, "fields.comment.comments[%d].author.name", i)) == nil)
			uid = "unknown";
		else
			adduser(s, uid);

		if((p = Jstrval(v, "fields.comment.comments[%d].created", i)) != nil)
			created = iso2sec(p);
		else
			created = time(nil);

		if((p = Jstrval(v, "fields.comment.comments[%d].updated", i)) != nil)
			updated = iso2sec(p);
		else
			updated = time(nil);

		if((p = Jstrval(v, "fields.comment.comments[%d].updateAuthor.name", i)) != nil)
			muid = p;
		else
			muid = uid;

		snprint(file, sizeof(file), "%d", i);
		/* abuse mtime and atime for created and updated (ugh) */
		fulltext(&n->child, file, uid, nil, muid, created, updated, body);
	}
}


static void
dumptree(Node *n, int in)
{

	for (; n; n = n->next){
		print("%*s %D\n", in, "", &n->dir);
		if (n->child)
			dumptree(n->child, in+2);
	}
}

static void
labels(Node **np, Jvalue *v)
{
	int i;
	Fmt f;
	char *p, *buf;

	fmtstrinit(&f);
	for(i = 0; (p = Jstrval(v, "fields.labels[%d]", i)) != nil; i++)
		fmtprint(&f, "%s\n", p);

	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	if(strlen(buf) > 0)
		subtext(np, "labels", buf);
	free(buf);
}


Node *
issueinfo_v5(Node *n, Sess *s, char *issue)
{
	Jvalue *v;
	Node *nn;
	char *p, *uid;

	if((v = get(s, "rest/api/%s/issue/%U", s->api, issue)) == nil){
		warn("%s cannot get issue detail - %r\n", issue);
		return n;
	}

	if((uid = Jstrval(v, "fields.reporter.name")) != nil){
		adduser(s, uid);
		fulltext(&n, "reporter", uid, nil, uid, 0, 0, uid);
	}
	else
		uid = "?";
	if((p = Jstrval(v, "fields.status.name")) != nil)
		fulltext(&n, "status", uid, nil, uid, 0, 0, p);
	if((p = Jstrval(v, "fields.resolution.name")) == nil)
		p = "Unresolved";
	fulltext(&n, "resolution", uid, nil, uid, 0, 0, p);
	if((p = Jstrval(v, "fields.workratio")) != nil)
		fulltext(&n, "workratio", uid, nil, uid, 0, 0, p);
	if((p = Jstrval(v, "fields.votes.votes")) != nil)
		fulltext(&n, "votes", uid, nil, uid, 0, 0, p);
	if((p = Jstrval(v, "fields.created")) != nil)
		fulltext(&n, "created", uid, nil, uid, 0, 0, ctime(iso2sec(p)));

	if((p = Jstrval(v, "fields.priority.name")) == nil)
		p = "";
	writehook(fulltext(&n, "priority", uid, nil, uid, 0, 0, p), upd_priority, issue);

	if((p = Jstrval(v, "fields.lastViewed")) != nil)
		fulltext(&n, "viewed", uid, nil, uid, 0, 0, ctime(iso2sec(p)));
	if((p = Jstrval(v, "fields.updated")) != nil)
		fulltext(&n, "updated", uid, nil, uid, 0, 0, ctime(iso2sec(p)));
	if((p = Jstrval(v, "fields.duedate")) != nil)
		fulltext(&n, "due", uid, nil, uid, 0, 0, ctime(iso2sec(p)));
	if((p = Jstrval(v, "fields.resolutiondate")) != nil)
		fulltext(&n, "resolved", uid, nil, uid, 0, 0, ctime(iso2sec(p)));
	if((p = Jstrval(v, "fields.environment")) != nil)
		fulltext(&n, "environment", uid, nil, uid, 0, 0, p);
	if((p = Jstrval(v, "FixVersions.name")) != nil)
		fulltext(&n, "fix-version", uid, nil, uid, 0, 0, p);
	if((p = Jstrval(v, "versions.name")) != nil)
		fulltext(&n, "affects-version", uid, nil, uid, 0, 0, p);

	if((p = Jstrval(v, "fields.assignee.name")) == nil)
		p = "";
	else
		adduser(s, p);
	writehook(fulltext(&n, "assignee", uid, nil, uid, 0, 0, p), upd_assignee, issue);

	if((p = Jstrval(v, "fields.issuetype.name")) == nil)
		p = "";
	writehook(fulltext(&n, "issuetype", uid, nil, uid, 0, 0, p), upd_issuetype, issue);

	if((p = Jstrval(v, "fields.summary")) == nil)
		p = "";
	writehook(fulltext(&n, "summary", uid, nil, uid, 0, 0, p), upd_summary, issue);

	if((p = Jstrval(v, "fields.description")) == nil)
		p = "";
	writehook(fulltext(&n, "description", uid, nil, uid, 0, 0, p), upd_description, issue);

	watchers(&n, v);
	versions(&n, v, "fixversions", "fixVersions");
	versions(&n, v, "affectsversions", "versions");
	components(&n, v);
	progress(&n, v, "progress.", "progress");
	progress(&n, v, "fields.aggregateprogress", "aggregateprogress");
	timetracking(&n, v);
	labels(&n, v);

	transitions(&n, s, issue);

	worklog(subdir(&n, nil, nil, "worklog"), v, s);
	subtasks(subdir(&n, nil, nil, "subtasks"), v);
	links(subdir(&n, nil, nil, "links"), v);

	nn = subdir(&n, nil, nil, "attachments");
	attachments(nn, v, s);
	createhook(nn, new_attachment, issue, 0);

	nn = subdir(&n, nil, nil, "comments");
	comments(nn, v, s);
	createhook(nn, new_comment, issue, 0);

	Jfree(v);
	return n;
}
