typedef struct Sess Sess;
typedef struct Node Node;

typedef Node *(*Getfunc)(Node *n, Sess *s, char *arg);
typedef int (*Putfunc)(Sess *s, char *arg, char *buf);

#pragma incomplete Node

struct Sess  {
	char *url;			/* host URL */
	int conid;			/* /net connection ID */
	char *webfs;		/* mountpoint of webfs */
	int fd;				/* webfs session */
	char *user;			/* who I authenticated as */
	char *api;			/* rest api version to use */
	int major;			/* major release version */
	int annon;			/* don't authenticate */
	Node **users;		/* root of /users directory */
	Node **groups;		/* root of /groups directory */
	Getfunc issueinfo;	/* correct issueinfo to call given server API version */
};

typedef struct Node Node;
struct Node {
	Node *parent;
	Node *child;
	Node *next;
	Dir dir;

	struct {			/* read file */
		char *buf;		/* buffer contains content */
		char *url;		/* url to load content from */
		long last;		/* time(2) contents generated */
	} read;

	struct {			/* scan dir */
		Getfunc func;	/* (re)load child node data */
		void *arg;		/* read arg */
		long last;		/* time(2) of last scan */
	} scan;

	struct {			/* walk to object */
		Getfunc func;	/* check file/dir exists */
		void *arg;		/* read arg */
		long last;		/* time(2) of last walk */
	} walk;

	struct {			/* write to file */
		Putfunc func;	/* write func */
		void *arg;		/* write arg */
	} write;

	struct {			/* creat file or dir */
		int isdir;		/* allowed to create a dir, not a file */
	} create;
};


#pragma	   varargck    type  "U"   char*
#pragma	   varargck    type  "E"   char*

extern int Dumpjson;
extern int Dumptree;
extern int Dumptime;
extern int Dumpcache;
extern int Dumpsend;
extern int Dumpvalfail;
extern long Userupdate;

/* get_common.c */
Node *uservalid(Node *n, Sess *s, char *user);
Node *userinfo(Node *n, Sess *s, char *user);
void adduser(Sess *s, char *uid);
Node *groupinfo(Node *n, Sess *s, char *group);
Node *groupvalid(Node *n, Sess *s, char *group);
void addgroup(Sess *s, char *uid);
Node *issuevalid(Node *n, Sess *s, char *issue);
Node *issuelist(Node *n, Sess *s, char *proj);
Node *projvalid(Node *n, Sess *s, char *proj);
Node *projinfo(Node *n, Sess *s, char *proj);
Node *queuelist(Node *n, Sess *s, char *user);
Node *resolvedlist(Node *n, Sess *s, char *user);
Node *reportedlist(Node *n, Sess *s, char *user);
Node *projlist(Node *n, Sess *s, char *);
Node *serverinfo(Node *n, Sess *s, char *);
Node *sessioninfo(Node *n, Sess *s, char *);
Node *globalinfo(Node *n, Sess *s, char *);

/* get_v4.c */
Node *issueinfo_v4(Node *n, Sess *s, char *issue);

/* get_v5.c */
Node *issueinfo_v5(Node *n, Sess *s, char *issue);

/* httperrstr.c */
void httperrstr(int n);

/* jira.c */
void warn(char *fmt, ...);
char *strlwr(char *str);
char *strupr(char *str);
Jvalue *get(Sess *s, char *fmt, ...);
int post(Sess *s, char *method, char *post, char *fmt, ...);
char *download(Sess *s, char *url, vlong length);
int upload(Sess *s, char *url, int src, vlong len);
int dialweb(Sess *s);
long iso2sec(char *s);

/* main.c */
void walkhook(Node *n, Getfunc func, char *arg);
void scanhook(Node *n, Getfunc func, char *arg);
void readhook(Node *n, Getfunc func, char *arg);
void writehook(Node *n, Putfunc func, char *arg);
void createhook(Node *n, Putfunc func, char *arg, int isdir);
Node *subdir(Node **np, char *uid, char *gid, char *name);
Node *fulltext(Node **np, char *name, char *uid, char *gid, char *muid, long mtime, long atime, char *text);
Node *fetchtext(Node **np, char *name, char *uid, char *gid, char *muid, long mtime, vlong size, char *url);
Node *subtext(Node **np, char *name, char *text);

/* put.c */
int upd_assignee(Sess *s, char *issue, char *str);
int upd_issuetype(Sess *s, char *issue, char *str);
int upd_priority(Sess *s, char *issue, char *str);
int upd_description(Sess *s, char *issue, char *str);
int upd_summary(Sess *s, char *issue, char *str);
int new_comment(Sess *s, char *issue, char *str);
int add_fixversion(Sess *s, char *issue, char *str);
int add_transition(Sess *s, char *issue, char *str);
int new_attachment(Sess *s, char *issue, char *str);
