/* ©  Copyright Steve Simon, 2013 */
#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <auth.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include <json.h>
#include "jira.h"

int
upd_assignee(Sess *s, char *issue, char *str)
{
	char buf[128], *p;

	if((p = strchr(str, '\n')) != nil)
		*p = 0;

	snprint(buf, sizeof(buf), "{ \"fields\" : { \"assignee\" : { \"name\" : \"%E\"} } }", p);
	return post(s, "PUT", buf, "rest/api/%s/issue/%U", s->api, issue);
}

int
upd_issuetype(Sess *s, char *issue, char *str)	/* does not work */
{
	char buf[128], *p;

	if((p = strchr(str, '\n')) != nil)
		*p = 0;

	snprint(buf, sizeof(buf), "{ \"fields\" : { \"issuetype\" : { \"name\" : \"%E\"} } }", p);
	return post(s, "PUT", buf, "rest/api/%s/issue/%U", s->api, issue);
}

int
upd_priority(Sess *s, char *issue, char *str)
{
	char buf[128], *p;

	if((p = strchr(str, '\n')) != nil)
		*p = 0;

	snprint(buf, sizeof(buf), "{ \"fields\" : { \"priority\" : { \"name\" : \"%s\"} } }", p);
	return post(s, "PUT", buf, "rest/api/%s/issue/%U", s->api, issue);
}

int
upd_description(Sess *s, char *issue, char *str)
{
	int rc;
	char *p;

	p = smprint("{ \"fields\" : { \"description\" : \"%E\" } }", str);
	rc = post(s, "PUT", p, "rest/api/%s/issue/%U", s->api, issue);
	free(p);
	return rc;
}

int
upd_summary(Sess *s, char *issue, char *str)
{
	int rc;
	char *p;

	p = smprint("{ \"fields\" : { \"summary\" : \"%E\" } }", str);
	rc = post(s, "PUT", p, "rest/api/%s/issue/%U", s->api, issue);
	free(p);
	return rc;
}

int
new_comment(Sess *s, char *issue, char *str)
{
	char *p;
	int rc;

	p = smprint("{ \"body\" : \"%E\"}", str);
	rc = post(s, "POST", p, "rest/api/%s/issue/%U/comment", s->api, issue);
	free(p);
	return rc;
}

int
new_attachment(Sess *, char *, char *)
{
	/* This is complex, we need a full MIME wrapper */
	return -1;
}

int
add_transition(Sess *s, char *issue, char *str)
{
	Fmt f;
	Jvalue *v;
	int i, rc, id, n;
	char *a[32], *name, *p, *buf, *sep, *val;

	if((n = getfields(str, a, nelem(a), 1, "\n")) < 1){
		werrstr("missing transition name");
		return -1;
	}

	if((v = get(s, "rest/api/%s/issue/%U/transitions?expand=transitions.fields", s->api, issue)) == nil){
		werrstr("cannot get valid transition list");
		return -1;
	}

	id = -1;
	for(i = 0; (name = Jstrval(v, "transitions[%d].name", i)) != nil; i++)
		if(cistrcmp(name, a[0]) == 0)
			if((p = Jstrval(v, "transitions[%d].id", i)) != nil)
				id = atoi(p);
	if(id == -1){
		werrstr("transition name not currently valid");
		return -1;
	}

	Jfree(v);

	fmtstrinit(&f);
	fmtprint(&f, "{ \"update\":");
	
	sep = "";
	for(i = 1; i < n; i++){
		if(a[i][0] == '\n' || a[1][0] == 0)
			continue;

		if((val = strchr(a[i], ':')) == nil){
			werrstr("%q field does not contain a ':'", a[i]);
			free(fmtstrflush(&f));
			return -1;
		}
		*val++ = 0;

		if(strcmp(a[i], "comment") == 0){
			fmtprint(&f, "%s { \"comment\": [ { \"add\": { \"body\": \"%E\" } } ] }", sep, val);
			sep=",";
			continue;
		}

		if(strcmp(a[i], "resolution") == 0){
			fmtprint(&f, "%s \"fields\": { \"resolution\": { \"name\": \"%E\" } } ", sep, val);
			sep=",";
			continue;
		}
	}
	fmtprint(&f, "%s \"transition\": { \"id\": \"%d\" } }", sep, id);

	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
fprint(2, "%s\n", buf);
	rc = post(s, "POST", buf, "rest/api/%s/issue/%U/transitions", s->api, issue);
	free(buf);
	return rc;
}

